/* Copyright 2014-2015, R.Harmsen. This source file is made available
   under the Creative Commons License Attribution-ShareAlike 4.0
   International, see http://creativecommons.org/licenses/by-sa/4.0/ .

   See also:
   http://rudhar.com/lingtics/intrlnga/cgi-grep/derautia.htm and
   http://rudhar.com/lingtics/intrlnga/cgi-grep/modempia.htm
   for clarifications.

   Program to convert dictionary information in
   Wikia format, like that in
   http://interlingua.wikia.com/wiki/Nederlandese-Interlingua
   http://interlingua.wikia.com/wiki/Interlingua-Nederlandese
   (but not
   http://interlingua.wikia.com/wiki/Germano-Interlingua
   as that is different and would make the program too complicated)
   to a format compatible with that by Paul Denisowsky:
   http://www.denisowski.org/Interlingua/IEDICT/iedict_readme.html
   (see also http://en.wikipedia.org/wiki/CEDICT ).

   Input (specified as call arguments: names of HTML files, one
   for each letter of the alphabet, as extracted (for example)
   with Unix 'fetch'.

   2 February 2014 (completed 3 Feb)

   10 November 2015
   Added support for grammatical categories coded as for example
   <i>v<br />
   <i>adj<br />
   instead of the more usual:
   <i>v</i>
   <i>adj</i>

   11-14 November 2015
   In case a vowel is underlined, indicating a deviant stress pattern,
   put the word in parentheses and repeat it before those, so the
   word will also be found with a simple search argument without the
   underlining.

   15 November 2015
   Added function PostfixBoldItalic .

   20 November 2015: Creative Commons
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static int DoFile (FILE *fp);
static int DoLine (char *buffr);
static int PrefixSplitUnderlines (char *s);
static int PostfixBoldItalic (char *s);
static int RepeatWordsWithUnderlining (char *s, size_t ssize);
static int ExpandIt (char *b, char *bb, char *e, char *ee, char *s, size_t ssize);

static long TotalLinesProcessed = 0;
static long LinesProcessed = 0;

static int verbose = 0;

int main (int argc, char **argv)
{
   FILE *fp;

   for (--argc, ++argv; argc > 0 && argv[0] && argv[0][0]; argc--, argv++)
   {
      if (argv[0][0] == '-')
      {
         /* Command line option */
         if (argv[0][1] == 'v')
         {
            verbose = 1;
         }
      }
      else
      {
         if ((fp = fopen(argv[0], "r")) == NULL)
         {
            fprintf(stderr, "Cannot open input file %s\n" , argv[0]);
            continue;
         }
         DoFile(fp);
         fprintf(stderr, "%ld lines processed from file %s\n",
            LinesProcessed, argv[0]);
         TotalLinesProcessed += LinesProcessed;
         LinesProcessed = 0;
      }
   }
   fprintf(stderr, "%ld lines processed in total\n", TotalLinesProcessed);

   return 0;
}

static char *StrtContentTrigger = "id=\"WikiaArticle\"";
static char *StopContentTrigger = "NewPP limit report";

static char buffr[1024];
static char Lemma[128];
static char Grcat[64];
static char SubLemma[256];
static char Entry[sizeof buffr - sizeof Lemma];
static char Outline[sizeof Lemma + sizeof Grcat + sizeof Entry + 10 + 5 + 5];

static int DoFile (FILE *fp)
{
   int inContent = 0;

   while (fgets(buffr, sizeof buffr, fp) != NULL)
   {
      /* Optimalisation: skip lots of Javascript etc. that is
         in the files before the actual dictionary entries */
      if (!inContent && strstr(buffr, StrtContentTrigger))
         inContent = 1;
      if (!inContent)
         continue;

      DoLine(buffr), LinesProcessed++;

      /* Yet another optimalisation: from a certain point in
         files, no more dictionary content follows, but the file
         hasn't ended by far. No need to wrestle through al that,
         better just skip it) */
      if (strstr(buffr, StopContentTrigger))
         break;
   }

   return 0;
}

static char *StrtLemma = "<b>";
static char *StopLemma = "</b>";

static char *StrtGrcat = "<i>";
static char *StopGrcat = "</i>";
static char *StopGrca2 = "<br />";

static char *StrtEntry = "<dl>";
static char *StopEntry = "</dl>";

static char *StrtExpre = "<b>";
static char *StrtExpr2 = "<dd>";
static char *StopExpre = "</b>";
static char *StopEntLn = "<br />";

static char *ReplMarkr = "--";

int inEntry = 0;
int inSubentry = 0;

static int DoLine (char *buffr)
{
   char *p, *q, *e;

   PrefixSplitUnderlines(buffr);

   if (!inEntry && (p = strstr(buffr, StrtEntry)) != NULL)
   {
      inEntry = 1;
   }
   if (inEntry && (p = strstr(buffr, StopEntry)) != NULL)
   {
      inEntry = 0;
   }

   if (!inEntry && (p = strstr(buffr, StrtLemma)) != NULL)
   {
      p += strlen(StrtLemma);
      q = strstr(p, StopLemma);

      if (q && q - p > sizeof Lemma - 1)
         q = p + sizeof Lemma - 1;

      if (q)
      {
         sprintf(Lemma, "%.*s", (int)(q - p), p);
      }
   }

   if (!inEntry && (p = strstr(buffr, StrtGrcat)) != NULL)
   {
      p += strlen(StrtGrcat);
      q = strstr(p, StopGrcat);
      /* Check also for things like 
         <i>v<br />
         <i>adj<br />
         which occur primarily under the letter p of Interlingua-Nederlands */
      if (!q)
         q = strstr(p, StopGrca2);

      if (q && q - p > sizeof Grcat - 1)
         q = p + sizeof Grcat - 1;

      if (q)
      {
         sprintf(Grcat, "%.*s", (int)(q - p), p);
      }
   }

   if (inEntry)
   {
      inSubentry = 0;
      if ((e = strchr(buffr, '=')) != NULL)
         inSubentry = 1;
      q = strstr(buffr, StopEntLn);

      if (!inSubentry)
      {
         p = strstr(buffr, StopLemma);
         if (p) p += strlen(StopLemma);
         while (p && *p == ' ')
            p++;

         if (p && q)
         {
            if (q - p > sizeof Entry - 1)
               q = p + sizeof Entry - 1;
            sprintf(Entry, "%.*s", (int)(q - p), p);

            sprintf(Outline, "%s /%s/ : %s", Lemma, Grcat, Entry);
            PostfixBoldItalic(Outline);
            RepeatWordsWithUnderlining(Outline, sizeof Outline);
            fprintf(stdout, "%s\n", Outline);
         }
      }
      else
      {
         char *qq;

         p = strstr(buffr, StrtExpre);
         if (p)
         {
            p += strlen(StrtExpre);
         }
         else
         {
            p = strstr(buffr, StrtExpr2);
            if (p)
            {
               p += strlen(StrtExpr2);
            }
         }

         q = strstr(buffr, StopExpre);
         /* There may be another stopper further on, then use that instead */
         while (q && (qq = strstr(q + 1, StopExpre)) && qq < e)
            q = qq;

         if (q)
         {
            char *r, *s;

            if (q - p > sizeof SubLemma - 1)
               q = p + sizeof SubLemma - 1;
            sprintf(SubLemma, "%.*s", (int)(q - p), p);

            p = e + 1;
            while (*p == ' ') p++;
            q = strstr(buffr, StopEntLn);
            if (q)
            {
               if (q - p > sizeof Entry - 1)
                  q = p + sizeof Entry - 1;
               sprintf(Entry, "%.*s", (int)(q - p), p);
            }

#if 0
            /* If the main lemma marker (e.g. -- ) is present in
               the sublemma, replace it with the actual main
               lemma. Note that the marker can occur more than once,
               hence (example: lemma 'toezicht'), hence the use of
               'for' instead of 'if'.

               After each replacement, don't start at the beginning
               again, but after the replacement, because the
               replacement marker sometimes also occurs in the main
               lemma. Examples:
               getroosten (zich --) WW
               Ardennen (de --) ZN EIGN MV

               5/6 February 2014:
               The replacement is not a good idea anyway, because it leads to
               incorrect Dutch. Examples:
               gaafe gulden (should be: gave), concreete vorm (should be: concrete),
               van boogen voorzien (should be: bogen), buisen (should be: buizen).

               Solution: don't do the expansion, but add the main lemma at the front,
               followed by a comma and space.
            */
            for (s = SubLemma;
                (r = strstr(s, ReplMarkr)) != NULL;
                 s = r + strlen(Lemma))
            {
               int needed = r - SubLemma;
               char *m;

               needed += strlen(Lemma) +
                         strlen(r + strlen(ReplMarkr)) + 1;
               if (!(m = malloc(needed)))
               {
                  fprintf(stderr, "Cannot malloc %d bytes\n", needed);
                  return 1;
               }
               sprintf(m, "%.*s%s%s",
                  (int)(r - SubLemma), SubLemma,
                  Lemma,
                  r + strlen(ReplMarkr));
               strncpy(SubLemma, m, sizeof SubLemma - 1);
               free(m);
            }
            fprintf(stdout, "%s : %s\n", SubLemma, Entry);
#else
            sprintf(Outline, "%s, %s : %s", Lemma, SubLemma, Entry);
            RepeatWordsWithUnderlining(Outline, sizeof Outline);
            PostfixBoldItalic(Outline);
            fprintf(stdout, "%s\n", Outline);
#endif
         }
      }
   }

   return 0;
}

/* If in Dutch a oo is stressed and must be indicated because there
   also a woord with a different stress (e.g. voorkmen and vrkomen),
   this is encoded not as <u>oo</u> but as <u>o</u><u>o</u>. As a
   result, the function RepeatWordsWithUnderlining does not properly
   handle them. To fix this the middle </u><u> is eliminated.

   The function also fixes some other anomalies.
 */
static char *DublUline = "</u><u>";
static char *Strange1 = "<i>c</i><b>o</b><i>n</i><b>j";
static char *Becomes1 = "<i>conj</i>";
static char *Strange2 = "<i>n</i><b>u</b><i>m</i><i>o</i><b>r</b><i>d</i>";
static char *Becomes2 = "<i>numord</i>";

static int PrefixSplitUnderlines (char *s)
{
   char *p;

   while ((p = strstr(s, DublUline)) != NULL)
   {
      memmove(p, p + strlen(DublUline), strlen(s) - (p - s) - strlen(DublUline));
   }
   if ((p = strstr(s, Strange1)) != NULL)
   {
      sprintf(p, "%s", Becomes1);
      memmove(p + strlen(Strange1), p + strlen(Becomes1), strlen(Strange1) - strlen(Becomes1));
   }
   if ((p = strstr(s, Strange2)) != NULL)
   {
      sprintf(p, "%s", Becomes2);
      memmove(p + strlen(Strange2), p + strlen(Becomes2), strlen(Strange2) - strlen(Becomes2));
   }

   return 0;
}

/* Sometimes a line somehow ends in bold or italic. Because
   line are later (cgi-grep.c) selected individually, that
   is undesirable. Blunt fix: add closing tags if that situation
   occurs. */
static int PostfixBoldItalic (char *s)
{
   char *p;
   int bold = 0;
   int italic = 0;

   for (p = s; *p; p++)
   {
      if (*p == '<')
      {
         if (p[1] == '/')
         {
            if (tolower(p[2]) == 'b')
               bold = 0;
            else if (tolower(p[2]) == 'i')
               italic = 0;
         }
         else
         {
            if (tolower(p[1]) == 'b')
               bold = 1;
            else if (tolower(p[1]) == 'i')
               italic = 1;
         }
      }

   }

   /* Insert closing tag just before newline. */
   if (bold)
   {
      if (verbose)
      {
         fprintf(stderr, "Extra </b> added to line:\n%s\n", s);
      }
      strcat(s, "</b>");
   }
   if (italic)
   {
      if (verbose)
      {
         fprintf(stderr, "Extra </i> added to line:\n%s\n", s);
      }
      strcat(s, "</i>");
   }

   return 0;
}

static char *StrUline = "<u>";
static char *EndUline = "</u>";

/* In case a vowel is underlined, indicating a deviant stress pattern,
   put the word in parentheses and repeat it before those, so the
   word will also be found with a simple search argument without the
   underlining. */
static int RepeatWordsWithUnderlining (char *s, size_t ssize)
{
   char *p, *b, *bb, *e, *ee, *q;
   size_t len;

   for (p = s; *p; )
   {
      int handled;

      b = strstr(p, StrUline);
      if (b && (e = strstr(b + strlen(StrUline), EndUline)))
      {
         /* Found an underlined string. */
         #if 0
         /* Now check if there are only vowels in between the tags.
            However, it's better not to check, because of the combi
            ij (which server as a diphthong in Dutch) and some errors
            in the dictionary (consonant before or after vowel
            underlined. */
         q = b + strlen(StrUline);
         if (len = strspn(q, "aeiouyAEIOUY"), len != 1 && len != 2)
         {
            p = e + strlen(EndUline);
            continue;
         }
         #endif

         bb = b;
         while (bb > p && (isalnum(bb[-1]) || bb[-1] == '-'))
         {
            bb--;
         }
         if (!isalnum(*bb) && *bb != '-' && *bb != '<')
            bb++;

         ee = e;
         ee +=  strlen(EndUline);
         while (isalnum(*ee) || *ee == '-')
            ee++;
         #ifdef DEBUG
            fprintf(stderr, "%4d: Found %*.*s\n", (int)(__LINE__), (int)(ee-bb), (int)(ee-bb), bb);
         #endif

         handled = ExpandIt(b, bb, e, ee, p, ssize - (p - s));
         if (handled <= 0)
            break;
         else
            p += handled;
      }
      else
      {
         /* No need to search for more in this line */
         break;
      }
   }

   return 0;
}

/* b points to start of underline tag <u>
   e points to start of underline end tag </u>
   bb points to start of word with underlined letter(s) in it
   ee points to just after word with underlined letter(s) in it

   Returns number of bytes written up until after the expansion and ). 
   Not counting what follows after that.
*/
static int ExpandIt (char *b, char *bb, char *e, char *ee, char *s, size_t ssize)
{
   size_t SpaceAvailable, ExtraSpaceNeeded; 
   char *wrk, *with, *without;
   /* Same pointers, but relative to buffer 'without' instead of to s */
   char *bw, *bbw, *ew, *eew;
   int handled;

   SpaceAvailable = ssize - strlen(s) - 1;
   ExtraSpaceNeeded = 3 /* space and 2 parentheses*/ +
      2 * (ee - bb) /* the word with and without the underlining tags */
      - strlen(StrUline) - strlen(EndUline);

   if (ExtraSpaceNeeded > SpaceAvailable)
   {
      /* Do nothing, safety first. Report situation to standard error. */
      fprintf(stderr, "%s line %d: LineOut = %s\nNeeded %d bytes, have only %d\n",
         __FILE__, (int)__LINE__,
         s, (int)ExtraSpaceNeeded, (int)SpaceAvailable);
      return -1;
   }

   if ((wrk = malloc(ssize)) == NULL)
   {
      fprintf(stderr, "%s line %d: Cannot malloc %d bytes as working buffer.\n",
         __FILE__, (int)__LINE__, (int)ssize);
      return -2;
   }

   if ((with = malloc(ee - bb + 1)) == NULL)
   {
      fprintf(stderr, "%s line %d: Cannot malloc %d bytes as working buffer.\n",
         __FILE__, (int)__LINE__, (int)(ee - bb + 1));
      free(wrk);
      return -3;
   }

   if ((without = malloc(ee - bb + 1)) == NULL)
   {
      fprintf(stderr, "%s line %d: Cannot malloc %d bytes as working buffer.\n",
         __FILE__, (int)__LINE__, (int)(ee - bb + 1));
      free(wrk); free(with);
      return -4;
   }
   sprintf(with, "%.*s", (int)(ee - bb), bb);
   strcpy(without, with);

   bw  = without + (b  - bb);
   ew  = without + (e  - bb);
   bbw = without + (bb - bb);
   eew = without + (ee - bb);
   memmove(ew, ew + strlen(EndUline), eew - ew - strlen(EndUline) + 1);
   memmove(bw, bw + strlen(StrUline), eew - bw - strlen(StrUline) - strlen(EndUline) + 1);

   handled = sprintf(wrk, "%.*s%s (%s)", (int)(bb - s), s, without, with);
   sprintf(wrk + handled, "%.*s", (int)(s + strlen(s) - ee), ee);
   #ifdef DEBUG
      fprintf(stderr, "%4d: with = %s, without = %s\n", (int)(__LINE__), with, without);
      fprintf(stderr, "%4d: bbw= %s, eew= %s\n", (int)(__LINE__), bb, ee);
      fprintf(stderr, "%4d: bw = %s, ew = %s\n", (int)(__LINE__), b , e );
      fprintf(stderr, "%4d: s   = %s\n", (int)(__LINE__), s);
      fprintf(stderr, "%4d: wrk = %s\n", (int)(__LINE__), wrk);
      fprintf(stderr, "%4d: handled bytes = %d\n\n", (int)(__LINE__), handled);
   #endif
   strcpy(s, wrk);

   free(wrk); free(with); free(without);

   return handled;
}
